@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2007-2015 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2007-2019 -- leonerd@leonerd.org.uk
 
 package IO::Async::Loop;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.71';
+our $VERSION = '0.74';
 
 # When editing this value don't forget to update the docs below
 use constant NEED_API_VERSION => '0.33';
@@ -758,7 +758,7 @@ invoked, in no particular order.
 The returned C<$id> value can be used to identify the signal handler in case
 it needs to be removed by the C<detach_signal> method. Note that this value
 may be an object reference, so if it is stored, it should be released after it
-cancelled, so the object itself can be freed.
+is cancelled, so the object itself can be freed.
 
 =over 8
 
@@ -888,95 +888,110 @@ sub later
    $loop->spawn_child( %params )
 
 This method creates a new child process to run a given code block or command.
-For more detail, see the C<spawn_child> method on the
-L<IO::Async::ChildManager> class.
-
-=cut
-
-sub spawn_child
-{
-   my $self = shift;
-   my %params = @_;
-
-   my $childmanager = $self->{childmanager} ||=
-      $self->__new_feature( "IO::Async::ChildManager" );
-
-   $childmanager->spawn_child( %params );
-}
-
-=head2 open_child
-
-   $pid = $loop->open_child( %params )
-
-This creates a new child process to run the given code block or command, and
-attaches filehandles to it that the parent will watch. This method is a light
-wrapper around constructing a new L<IO::Async::Process> object, provided
-largely for backward compatibility. New code ought to construct such an object
-directly, as it may provide more features than are available here.
-
 The C<%params> hash takes the following keys:
 
 =over 8
 
 =item command => ARRAY or STRING
 
+Either a reference to an array containing the command and its arguments, or a
+plain string containing the command. This value is passed into perl's
+C<exec> function.
+
 =item code => CODE
 
-The command or code to run in the child process (as per the C<spawn> method)
-
-=item on_finish => CODE
-
-A continuation to be called when the child process exits and has closed all of
-the filehandles that were set up for it. It will be invoked in the following
-way:
-
- $on_finish->( $pid, $exitcode )
-
-The second argument is passed the plain perl C<$?> value.
-
-=item on_error => CODE
-
-Optional continuation to be called when the child code block throws an
-exception, or the command could not be C<exec(2)>ed. It will be invoked in the
-following way (as per C<spawn>)
-
- $on_error->( $pid, $exitcode, $dollarbang, $dollarat )
-
-If this continuation is not supplied, then C<on_finish> is used instead. The
-value of C<$!> and C<$@> will not be reported.
+A block of code to execute in the child process. It will be called in scalar
+context inside an C<eval> block.
 
 =item setup => ARRAY
 
-Optional reference to an array to pass to the underlying C<spawn> method.
+A reference to an array which gives file descriptors to set up in the child
+process before running the code or command. See below.
+
+=item on_exit => CODE
+
+A continuation to be called when the child processes exits. It will be invoked
+in the following way:
+
+ $on_exit->( $pid, $exitcode, $dollarbang, $dollarat )
+
+The second argument is passed the plain perl C<$?> value.
 
 =back
 
-In addition, the hash takes keys that define how to set up file descriptors in
-the child process. (If the C<setup> array is also given, these operations will
-be performed after those specified by C<setup>.)
+Exactly one of the C<command> or C<code> keys must be specified.
+
+If the C<command> key is used, the given array or string is executed using the
+C<exec> function. 
+
+If the C<code> key is used, the return value will be used as the C<exit(2)>
+code from the child if it returns (or 255 if it returned C<undef> or thows an
+exception).
+
+ Case          | ($exitcode >> 8)       | $dollarbang | $dollarat
+ --------------+------------------------+-------------+----------
+ exec succeeds | exit code from program |     0       |    ""
+ exec fails    |         255            |     $!      |    ""
+ $code returns |     return value       |     $!      |    ""
+ $code dies    |         255            |     $!      |    $@
+
+It is usually more convenient to use the C<open_process> method in simple
+cases where an external program is being started in order to interact with it
+via file IO, or even C<run_child> when only the final result is required,
+rather than interaction while it is running.
+
+=head3 C<setup> array
+
+This array gives a list of file descriptor operations to perform in the child
+process after it has been C<fork(2)>ed from the parent, before running the code
+or command. It consists of name/value pairs which are ordered; the operations
+are performed in the order given.
 
 =over 8
 
-=item fdI<n> => HASH
+=item fdI<n> => ARRAY
 
-A hash describing how to set up file descriptor I<n>. The hash may contain one
-of the following sets of keys:
+Gives an operation on file descriptor I<n>. The first element of the array
+defines the operation to be performed:
 
 =over 4
 
-=item on_read => CODE
+=item [ 'close' ]
 
-The child will be given the writing end of a pipe. The reading end will be
-wrapped by an L<IO::Async::Stream> using this C<on_read> callback function.
+The file descriptor will be closed.
 
-=item from => STRING
+=item [ 'dup', $io ]
 
-The child will be given the reading end of a pipe. The string given by the
-C<from> parameter will be written to the child. When all of the data has been
-written the pipe will be closed.
+The file descriptor will be C<dup2(2)>ed from the given IO handle.
+
+=item [ 'open', $mode, $file ]
+
+The file descriptor will be opened from the named file in the given mode. The
+C<$mode> string should be in the form usually given to the C<open> function;
+such as '<' or '>>'.
+
+=item [ 'keep' ]
+
+The file descriptor will not be closed; it will be left as-is.
 
 =back
 
+A non-reference value may be passed as a shortcut, where it would contain the
+name of the operation with no arguments (i.e. for the C<close> and C<keep>
+operations).
+
+=item IO => ARRAY
+
+Shortcut for passing C<fdI<n>>, where I<n> is the fileno of the IO
+reference. In this case, the key must be a reference that implements the
+C<fileno> method. This is mostly useful for
+
+ $handle => 'keep'
+
+=item fdI<n> => IO
+
+A shortcut for the C<dup> case given above.
+
 =item stdin => ...
 
 =item stdout => ...
@@ -985,8 +1000,154 @@ written the pipe will be closed.
 
 Shortcuts for C<fd0>, C<fd1> and C<fd2> respectively.
 
+=item env => HASH
+
+A reference to a hash to set as the child process's environment.
+
+Note that this will entirely set a new environment, completely replacing the
+existing one. If you want to simply add new keys or change the values of some
+keys without removing the other existing ones, you can simply copy C<%ENV>
+into the hash before setting new keys:
+
+ env => {
+    %ENV,
+    ANOTHER => "key here",
+ }
+
+=item nice => INT
+
+Change the child process's scheduling priority using C<POSIX::nice>.
+
+=item chdir => STRING
+
+Change the child process's working directory using C<chdir>.
+
+=item setuid => INT
+
+=item setgid => INT
+
+Change the child process's effective UID or GID.
+
+=item setgroups => ARRAY
+
+Change the child process's groups list, to those groups whose numbers are
+given in the ARRAY reference.
+
+On most systems, only the privileged superuser change user or group IDs.
+L<IO::Async> will B<NOT> check before detaching the child process whether
+this is the case.
+
+If setting both the primary GID and the supplementary groups list, it is
+suggested to set the primary GID first. Moreover, some operating systems may
+require that the supplementary groups list contains the primary GID.
+
 =back
 
+If no directions for what to do with C<stdin>, C<stdout> and C<stderr> are
+given, a default of C<keep> is implied. All other file descriptors will be
+closed, unless a C<keep> operation is given for them.
+
+If C<setuid> is used, be sure to place it after any other operations that
+might require superuser privileges, such as C<setgid> or opening special
+files.
+
+Z<>
+
+   my ( $pipeRd, $pipeWr ) = IO::Async::OS->pipepair;
+   $loop->spawn_child(
+      command => "/usr/bin/my-command",
+
+      setup => [
+         stdin  => [ "open", "<", "/dev/null" ],
+         stdout => $pipeWr,
+         stderr => [ "open", ">>", "/var/log/mycmd.log" ],
+         chdir  => "/",
+      ]
+
+      on_exit => sub {
+         my ( $pid, $exitcode ) = @_;
+         my $status = ( $exitcode >> 8 );
+         print "Command exited with status $status\n";
+      },
+   );
+
+   $loop->spawn_child(
+      code => sub {
+         do_something; # executes in a child process
+         return 1;
+      },
+
+      on_exit => sub {
+         my ( $pid, $exitcode, $dollarbang, $dollarat ) = @_;
+         my $status = ( $exitcode >> 8 );
+         print "Child process exited with status $status\n";
+         print " OS error was $dollarbang, exception was $dollarat\n";
+      },
+   );
+
+=cut
+
+sub spawn_child
+{
+   my $self = shift;
+   my %params = @_;
+
+   my $childmanager = $self->{childmanager} ||=
+      $self->__new_feature( "IO::Async::Internals::ChildManager" );
+
+   $childmanager->spawn_child( %params );
+}
+
+=head2 open_process
+
+   $process = $loop->open_process( %params )
+
+I<Since version 0.72.>
+
+This creates a new child process to run the given code block or command, and
+attaches filehandles to it that the parent will watch. This method is a light
+wrapper around constructing a new L<IO::Async::Process> object, adding it to
+the loop, and returning it.
+
+The C<%params> hash is passed directly to the L<IO::Async::Process>
+constructor.
+
+=cut
+
+sub open_process
+{
+   my $self = shift;
+   my %params = @_;
+
+   $params{on_exit} and croak "Cannot pass 'on_exit' parameter through ->open_process";
+
+   require IO::Async::Process;
+   my $process = IO::Async::Process->new( %params );
+
+   $self->add( $process );
+
+   return $process;
+}
+
+=head2 open_child
+
+   $pid = $loop->open_child( %params )
+
+A back-compatibility wrapper to calling L</open_process> and returning the PID
+of the newly-constructed L<IO::Async::Process> instance. The C<on_finish>
+continuation likewise will be invoked with the PID rather than the process
+instance.
+
+   $on_finish->( $pid, $exitcode )
+
+Similarly, a C<on_error> continuation is accepted, though note its arguments
+come in a different order to those of the Process's C<on_exception>:
+
+   $on_error->( $pid, $exitcode, $errno, $exception )
+
+This method should not be used in new code; instead use L</open_process>
+directly.
+
 =cut
 
 sub open_child
@@ -1011,23 +1172,21 @@ sub open_child
       };
    }
 
-   $params{on_exit} and croak "Cannot pass 'on_exit' parameter through ChildManager->open";
-
-   require IO::Async::Process;
-   my $process = IO::Async::Process->new( %params );
-
-   $self->add( $process );
-
-   return $process->pid;
+   return $self->open_process( %params )->pid;
 }
 
-=head2 run_child
+=head2 run_process
 
-   $pid = $loop->run_child( %params )
+   @results = $loop->run_process( %params )->get
 
-This creates a new child process to run the given code block or command,
-capturing its STDOUT and STDERR streams. When the process exits, a
-continuation is invoked being passed the exitcode, and content of the streams.
+   ( $exitcode, $stdout ) = $loop->run_process( ... )->get  # by default
+
+I<Since version 0.73.>
+
+Creates a new child process to run the given code block or command, optionally
+capturing its STDOUT and STDERR streams. By default the returned future will
+yield the exit code and content of the STDOUT stream, but the C<capture>
+argument can be used to alter what is requested and returned.
 
 =over 8
 
@@ -1038,6 +1197,143 @@ continuation is invoked being passed the exitcode, and content of the streams.
 The command or code to run in the child process (as per the C<spawn_child>
 method)
 
+=item stdin => STRING
+
+Optional. String to pass in to the child process's STDIN stream.
+
+=item setup => ARRAY
+
+Optional reference to an array to pass to the underlying C<spawn> method.
+
+=item capture => ARRAY
+
+Optional reference to an array giving a list of names of values which should
+be returned by resolving future. Values will be returned in the same order as
+in the list. Valid choices are: C<exitcode>, C<stdout>, C<stderr>.
+
+=item cancel_signal => STRING
+
+Optional. Name (or number) of the signal to send to the process if the
+returned future is cancelled. Defaults to C<TERM>. Use empty string or zero
+disable sending a signal on cancellation.
+
+=item fail_on_nonzero => BOOL
+
+Optional. If true, the returned future will fail if the process exits with a
+nonzero status. The failure will contain a message, the C<process> category
+name, and the capture values that were requested.
+
+   Future->fail( $message, process => @captures )
+
+=back
+
+This method is intended mainly as an IO::Async-compatible replacement for the
+perl C<readpipe> function (`backticks`), allowing it to replace
+
+   my $output = `command here`;
+
+with
+
+   my ( $exitcode, $output ) = $loop->run_process(
+      command => "command here", 
+   )->get;
+
+Z<>
+
+   my ( $exitcode, $stdout ) = $loop->run_process(
+      command => "/bin/ps",
+   )->get;
+
+   my $status = ( $exitcode >> 8 );
+   print "ps exited with status $status\n";
+
+=cut
+
+sub _run_process
+{
+   my $self = shift;
+   my %params = @_;
+
+   $params{on_finish} and croak "Unrecognised parameter on_finish";
+
+   my $capture = delete $params{capture} // [qw(exitcode stdout)];
+   ref $capture eq "ARRAY" or croak "Expected 'capture' to be an array reference";
+
+   my %subparams;
+   my %results;
+
+   if( my $child_stdin = delete $params{stdin} ) {
+      ref $child_stdin and croak "Expected 'stdin' not to be a reference";
+      $subparams{stdin} = { from => $child_stdin };
+   }
+
+   $subparams{code}    = delete $params{code};
+   $subparams{command} = delete $params{command};
+   $subparams{setup}   = delete $params{setup};
+
+   foreach my $name ( @$capture ) {
+      grep { $_ eq $name } qw( exitcode stdout stderr ) or croak "Unexpected capture $name";
+
+      $subparams{stdout} = { into => \$results{stdout} } if $name eq "stdout";
+      $subparams{stderr} = { into => \$results{stderr} } if $name eq "stderr";
+   }
+
+   my $cancel_signal = delete $params{cancel_signal} // "TERM";
+
+   my $fail_on_nonzero = delete $params{fail_on_nonzero};
+
+   croak "Unrecognised parameters " . join( ", ", keys %params ) if keys %params;
+
+   my $future = $self->new_future;
+
+   require IO::Async::Process;
+   my $process = IO::Async::Process->new(
+      %subparams,
+      on_finish => sub {
+         ( undef, $results{exitcode} ) = @_;
+
+         if( $fail_on_nonzero and $results{exitcode} > 0 ) {
+            $future->fail( "Process failed with exit code $results{exitcode}\n",
+               process => @results{ @$capture }
+            );
+         }
+         else {
+            $future->done( @results{ @$capture } );
+         }
+      },
+   );
+
+   $future->on_cancel(sub {
+      $process->kill( $cancel_signal );
+   }) if $cancel_signal;
+
+   $self->add( $process );
+
+   return ( $future, $process );
+}
+
+sub run_process
+{
+   my $self = shift;
+   return ( $self->_run_process( @_ ) )[0];
+}
+
+=head2 run_child
+
+   $pid = $loop->run_child( %params )
+
+A back-compatibility wrapper for L</run_process>, returning the PID and taking
+an C<on_finish> continuation instead of returning a Future.
+
+This creates a new child process to run the given code block or command,
+capturing its STDOUT and STDERR streams. When the process exits, a
+continuation is invoked being passed the exitcode, and content of the streams.
+
+Takes the following named arguments in addition to those taken by
+C<run_process>:
+
+=over 8
+
 =item on_finish => CODE
 
 A continuation to be called when the child process exits and closed its STDOUT
@@ -1047,30 +1343,10 @@ and STDERR streams. It will be invoked in the following way:
 
 The second argument is passed the plain perl C<$?> value.
 
-=item stdin => STRING
-
-Optional. String to pass in to the child process's STDIN stream.
-
-=item setup => ARRAY
-
-Optional reference to an array to pass to the underlying C<spawn> method.
-
 =back
 
-This method is intended mainly as an IO::Async-compatible replacement for the
-perl C<readpipe> function (`backticks`), allowing it to replace
-
-  my $output = `command here`;
-
-with
-
- $loop->run_child(
-    command => "command here", 
-    on_finish => sub {
-       my ( undef, $exitcode, $output ) = @_;
-       ...
-    }
- );
+This method should not be used in new code; intead use L</run_process>
+directly.
 
 =cut
 
@@ -1082,37 +1358,18 @@ sub run_child
    my $on_finish = delete $params{on_finish};
    ref $on_finish or croak "Expected 'on_finish' to be a reference";
 
-   my $stdout;
-   my $stderr;
-
-   my %subparams;
-
-   if( my $child_stdin = delete $params{stdin} ) {
-      ref $child_stdin and croak "Expected 'stdin' not to be a reference";
-      $subparams{stdin} = { from => $child_stdin };
-   }
-
-   $subparams{code}    = delete $params{code};
-   $subparams{command} = delete $params{command};
-   $subparams{setup}   = delete $params{setup};
-
-   croak "Unrecognised parameters " . join( ", ", keys %params ) if keys %params;
-
-   require IO::Async::Process;
-   my $process = IO::Async::Process->new(
-      %subparams,
-      stdout => { into => \$stdout },
-      stderr => { into => \$stderr },
-
-      on_finish => sub {
-         my ( $process, $exitcode ) = @_;
-         $on_finish->( $process->pid, $exitcode, $stdout, $stderr );
-      },
+   my ( $f, $process ) = $self->_run_process(
+      %params,
+      capture => [qw( exitcode stdout stderr )],
    );
+   my $pid = $process->pid;
 
-   $self->add( $process );
+   $f->on_done( sub {
+      undef $f; # capture cycle
+      $on_finish->( $pid, @_ );
+   });
 
-   return $process->pid;
+   return $pid;
 }
 
 =head2 resolver
@@ -1783,6 +2040,8 @@ sub _listen_addrs
          next;
       }
 
+      $sock->blocking( 0 );
+
       if( $reuseaddr ) {
          unless( $sock->sockopt( SO_REUSEADDR, 1 ) ) {
             $sockopterr = $!;
@@ -1940,6 +2199,10 @@ sub fork
          }
       }
 
+      # If the child process wants to use an IO::Async::Loop it needs to make
+      # a new one, so this value is never useful
+      undef our $ONE_TRUE_LOOP;
+
       my $exitvalue = eval { $code->() };
 
       defined $exitvalue or $exitvalue = -1;
@@ -2016,6 +2279,7 @@ sub create_thread
    unless( $self->{thread_join_pipe} ) {
       ( my $rd, $self->{thread_join_pipe} ) = IO::Async::OS->pipepair or
          croak "Cannot pipepair - $!";
+      $rd->blocking( 0 );
       $self->{thread_join_pipe}->autoflush(1);
 
       $self->watch_io(
